/* The standard CSS for doxygen */

body, table, div, p, dl {
  font-family: Lucida Grande, Verdana, Geneva, Arial, sans-serif;
  font-size: 12px;
}

/* @group Heading Levels */

h1 {
  font-size: 150%;
  color: #ffffff;
}

.title {
  font-size: 150%;
  font-weight: bold;
  margin: 10px 2px;
  color: #ffffff;
}

h2 {
  font-size: 120%;
  color: #ffffff;
}

h3 {
  font-size: 100%;
  color: #ffffff;
}

dt {
  font-weight: bold;
}

div.multicol {
  -moz-column-gap: 1em;
  -webkit-column-gap: 1em;
  -moz-column-count: 3;
  -webkit-column-count: 3;
}

p.startli, p.startdd, p.starttd {
  margin-top: 2px;
}

p.endli {
  margin-bottom: 0px;
}

p.enddd {
  margin-bottom: 4px;
}

p.endtd {
  margin-bottom: 2px;
}

/* @end */

caption {
  font-weight: bold;
}

span.legend {
  font-size: 70%;
  text-align: center;
}

h3.version {
  font-size: 90%;
  text-align: center;
}

div.qindex, div.navtab{
  background-color: #000000;
  border: 1px solid #333333;
  text-align: center;
  margin: 2px;
  padding: 2px;
}

div.qindex, div.navpath {
  width: 100%;
  line-height: 140%;
}

div.navtab {
  margin-right: 15px;
}

/* @group Link Styling */

a {
  color: #BB2222;
  font-weight: normal;
  text-decoration: none;
}

.contents a:visited {
  color: #BB2222;
}

a:hover {
  text-decoration: underline;
}

a.qindex {
  font-weight: bold;
}

a.qindexHL {
  font-weight: bold;
  background-color: #9CAFD4;
  color: #ffffff;
  border: 1px double #869DCA;
}

.contents a.qindexHL:visited {
  color: #ffffff;
}

a.el {
  font-weight: bold;
}

a.elRef {

}

a.code {
  color: #BB2222;
}

a.codeRef {
  color: #BB2222;
}

/* @end */

dl.el {
  margin-left: -1cm;
}

.fragment {
  font-family: monospace, fixed;
  font-size: 105%;
}

pre.fragment {
  border: 5px solid #1D1D1D;
  background-color: #2D2D2D;
  padding: 10px 10px 10px 10px;
  page-break-before: avoid;
  overflow: auto;
  word-wrap: break-word;
  font-size: 90%;
  margin-left: 1.75em;
  margin-right: 1.75em;
  margin-top: 1em;
  margin-bottom: 1em;
  color: #ffffff;
}

div.ah {
  background-color: black;
  font-weight: bold;
  color: #ffffff;
  margin-bottom: 3px;
  margin-top: 3px;
  padding: 0.2em;
  border: solid thin #333;
}

div.groupHeader {
  margin-left: 16px;
  margin-top: 12px;
  font-weight: bold;
}

div.groupText {
  margin-left: 16px;
  font-style: italic;
}

body {
  background: #000000;
  color: #808080;
  margin: 0;
}

div.contents {
  margin-top: 10px;
  margin-left: 10px;
  margin-right: 5px;
}

td.indexkey {
  background-color: #000000;
  font-weight: bold;
  border: 1px solid #333333;
  margin: 2px 0px 2px 0;
  padding: 2px 10px;
}

td.indexvalue {
  background-color: #000000;
  border: 1px solid #333333;
  padding: 2px 10px;
  margin: 2px 0px;
}

tr.memlist {
  background-color: #EEF1F7;
}

p.formulaDsp {
  text-align: center;
}

img.formulaDsp {

}

img.formulaInl {
  vertical-align: middle;
}

div.center {
  text-align: center;
  margin-top: 0px;
  margin-bottom: 0px;
  padding: 0px;
}

div.center img {
  border: 0px;
}

address.footer {
  text-align: right;
  padding-right: 12px;
}

img.footer {
  border: 0px;
  vertical-align: middle;
}

/* @group Code Colorization */

span.keyword {
  color: #FF0000;
}

span.keywordtype {
  color: #FF00FF;
}

span.keywordflow {
  color: #800080;
}

span.comment {
  color: #00FFFF;
}

span.preprocessor {
  color: #808080;
}

span.stringliteral {
  color: #FFFF00;
}

span.charliteral {
color: #FFFF00;
}

span.vhdldigit {
  color: #FFFF00;
}

span.vhdlchar {
  color: #FFFF00;
}

span.vhdlkeyword {
  color: #FF0000;
}

span.vhdllogic {
  color: #FF0000;
}

/* @end */

/*
   .search {
color: #003399;
font-weight: bold;
}

form.search {
margin-bottom: 0px;
margin-top: 0px;
}

input.search {
font-size: 75%;
color: #000080;
font-weight: normal;
background-color: #e8eef2;
}
 */

td.tiny {
  font-size: 75%;
}

.dirtab {
  padding: 4px;
  border-collapse: collapse;
  border: 1px solid #A3B4D7;
}

th.dirtab {
  background: #EBEFF6;
  font-weight: bold;
}

hr {
  height: 0px;
  border: none;
  border-top: 3px solid #BB2222;
}

hr.footer {
  height: 1px;
}

/* @group Member Descriptions */

table.memberdecls {
  border-spacing: 0px;
  padding: 0px;
}

.mdescLeft, .mdescRight,
.memItemLeft, .memItemRight,
.memTemplItemLeft, .memTemplItemRight, .memTemplParams {
  background-color: #000000;
  border: none;
  margin: 4px;
  padding: 1px 0 0 8px;
}

.mdescLeft, .mdescRight {
  padding: 0px 8px 4px 8px;
  color: #555;
}

.memItemLeft, .memItemRight, .memTemplParams {
  border-top: 1px solid #333333;
}

.memItemLeft, .memTemplItemLeft {
  white-space: nowrap;
}

.memItemRight {
  width: 100%;
}

.memTemplParams {
  color: #FFFFFF;
  white-space: nowrap;
}

/* @end */

/* @group Member Details */

/* Styles for detailed member documentation */

.memtemplate {
  color: #FFFFFF;
  font-weight: bold;
  margin-left: 8px;
  font-family: Andalo Mono, Courier New, Courier, Lucida Typewrite, fixed;
}

.memnav {
  background-color: #000000;
  border: 1px solid #333333;
  text-align: center;
  margin: 2px;
  margin-right: 15px;
  padding: 2px;
}

.mempage {
  width: 100%;
}

.memitem {
  padding: 0;
  margin-bottom: 10px;
  margin-right: 5px;
}

.memname {
  white-space: nowrap;
  font-weight: bold;
  margin-left: 6px;
  font-family: Andale Mono, Courier New, Courier, Lucida Typewriter, fixed;
}

.memproto {
  border-top: 1px solid #808080;
  border-left: 1px solid #808080;
  border-right: 1px solid #808080;
  padding: 6px 0px 6px 0px;
  color: #FFFFFF;
  font-weight: bold;
}

.memdoc {
  border-bottom: 1px solid #808080;
  border-left: 1px solid #808080;
  border-right: 1px solid #808080;
  border-top: 1px solid #333333;
  padding: 2px 5px;
}

.paramkey {
  text-align: right;
}

.paramtype {
  white-space: nowrap;
  color: #808080;
  font-family: Andale Mono, Courier New, Courier, Lucida Typewriter, fixed;
}

.paramname {
  color: #BB2222;
  white-space: nowrap;
  font-family: Andale Mono, Courier New, Courier, Lucida Typewriter, fixed;
}

.paramname em {
  font-style: normal;
}

.params, .retval, .exception, .tparams {
  border-spacing: 6px 2px;
}

.params .paramname, .retval .paramname {
  font-weight: bold;
  vertical-align: top;
}

.params .paramtype {
  font-style: italic;
  vertical-align: top;
}

.params .paramdir {
  font-family: "courier new",courier,monospace;
  vertical-align: top;
}

/* @end */

/* @group Directory (tree) */

/* for the tree view */

.ftvtree {
  font-family: sans-serif;
  margin: 0px;
}

/* these are for tree view when used as main index */

.directory {
  font-size: 9pt;
  font-weight: bold;
  margin: 5px;
}

.directory h3 {
  margin: 0px;
  margin-top: 1em;
  font-size: 11pt;
}

/*
   The following two styles can be used to replace the root node title
   with an image of your choice.  Simply uncomment the next two styles,
   specify the name of your image and be sure to set 'height' to the
   proper pixel height of your image.
 */

/*
   .directory h3.swap {
height: 61px;
background-repeat: no-repeat;
background-image: url("yourimage.gif");
}
.directory h3.swap span {
display: none;
}
 */

.directory > h3 {
  margin-top: 0;
}

.directory p {
  margin: 0px;
  white-space: nowrap;
}

.directory div {
  display: none;
  margin: 0px;
}

.directory img {
  vertical-align: -30%;
}

/* these are for tree view when not used as main index */

.directory-alt {
  font-size: 100%;
  font-weight: bold;
}

.directory-alt h3 {
  margin: 0px;
  margin-top: 1em;
  font-size: 11pt;
}

.directory-alt > h3 {
  margin-top: 0;
}

.directory-alt p {
  margin: 0px;
  white-space: nowrap;
}

.directory-alt div {
  display: none;
  margin: 0px;
}

.directory-alt img {
  vertical-align: -30%;
}

/* @end */

div.dynheader {
  margin-top: 8px;
}

address {
  font-style: normal;
  color: #2A3D61;
}

table.doxtable {
  border-collapse: collapse;
}

table.doxtable td, table.doxtable th {
  border: 1px solid #2D4068;
  padding: 3px 7px 2px;
}

table.doxtable th {
  background-color: #374F7F;
  color: #FFFFFF;
  font-size: 110%;
  padding-bottom: 4px;
  padding-top: 5px;
  text-align: left;
}

.tabsearch {
  top: 0px;
  left: 10px;
  height: 36px;
  background-image: url('tab_b.png');
  z-index: 101;
  overflow: hidden;
  font-size: 13px;
}

.navpath ul {
  font-size: 11px;
  background: #000000;
  color: #8AA0CC;
  border-bottom: 1px solid #333333;
  overflow: hidden;
  margin: 0px;
  padding-top: 0.25em;
  padding-bottom: 0.25em;
  padding-left: 0.5em;
  padding-right: 0;
  border-left: 1px solid #333333;
}

.navpath li {
  list-style-type: none;
  float: left;
  padding-right: 0.5em;
  color: #364D7C;
  border-right: 1px solid #333333;
  padding-left: 0.5em;
}

.navpath li.navelem a {
  display: block;
  text-decoration: none;
  outline: none;
}

.navpath li.navelem a:hover {
  color:#FFFFFF;
}

.navpath li.footer {
  list-style-type: none;
  float: right;
  padding-left: 10px;
  padding-right: 15px;
  background-image: none;
  background-repeat: no-repeat;
  background-position: right;
  color: #364D7C;
  font-size: 8pt;
}

div.summary {
  float: right;
  font-size: 8pt;
  padding-right: 5px;
  width: 50%;
  text-align: right;
}

div.summary a {
  white-space: nowrap;
}

div.ingroups {
  font-size: 8pt;
  padding-left: 5px;
  width: 50%;
  text-align: left;
}

div.ingroups a {
  white-space: nowrap;
}

div.header {
  background-color: #000000;
  margin: 0px;
  border-bottom: 1px solid #333333;
}

div.headertitle {
  padding: 5px 5px 5px 10px;
}

dl {
  padding: 0 0 0 10px;
}

dl.note, dl.warning, dl.attention, dl.pre, dl.post, dl.invariant,
dl.deprecated, dl.todo, dl.test, dl.bug {
  border-left: 4px solid;
  padding: 0 0 0 6px;
}

dl.note {
  border-color: #D0C000;
}

dl.warning, dl.attention {
  border-color: #FF0000;
}

dl.pre, dl.post, dl.invariant {
  border-color: #00D000;
}

dl.deprecated {
  border-color: #505050;
}

dl.todo {
  border-color: #00C0E0;
}

dl.test {
  border-color: #3030E0;
}

dl.bug {
  border-color: #C08050;
}

#projectlogo {
  text-align: center;
  vertical-align: bottom;
  border-collapse: separate;
}

#projectlogo img {
  border: 0px none;
}

#projectname {
  font: 300% Tahoma, Arial, sans-serif;
  margin: 0px;
  padding: 2px 0px;
}

#projectbrief {
  font: 120% Tahoma, Arial, sans-serif;
  margin: 0px;
  padding: 0px;
}

#projectnumber {
  font: 50% Tahoma, Arial,sans-serif;
  margin: 0px;
  padding: 0px;
}

#titlearea {
  padding: 0px;
  margin: 0px;
  width: 100%;
  border-bottom: 1px solid #808080;
}

.image {
  text-align: center;
}

.dotgraph {
  text-align: center;
}

.mscgraph {
  text-align: center;
}

.caption {
  font-weight: bold;
}

/** tab list at top of page */
.tabs, .tabs2, .tabs3 {
  background-image: none !important;
  background: #000000;
  border-left: 1px solid #333333;
  border-right: 1px solid #333333;
  border-bottom: 1px solid #333333;
  min-height: 1.5em;
}

.tablist li {
  background-image: none !important;
  background: #000000;
  border-right: 1px solid #333333;
  height: auto !important;
  padding-bottom: 0.25em;
  padding-top: 0.25em;
  line-height: 1em !important;
}

.tablist li.current {
  background: #BB2222;
}

.tablist li.current a {
  background-image: none !important;
  text-shadow: none;
  color: #ffffff;
}

.tablist a {
  background-image: none !important;
  text-shadow: none;
  color: #ffffff;
  font-weight: bold;
}

.tablist li:hover {
  background: #333333;
}

.tablist li.current:hover {
  background: #BB2222 !important;
}

/***
 * For trac-doxygen; these rules won't apply otherwise.
 */
div.tabs span {
  background-image: none !important;
  background: transparent !important;
  height: auto !important;
  padding-bottom: 0.25em;
  padding-top: 0.25em;
  line-height: 1em !important;
}

div.tabs a {
  background-image: none !important;
  background: transparent !important;
  border-bottom: none !important;
  font-size: 100% !important;
}

div.tabs span {
  padding-bottom: 0.25em;
  padding-top: 0.25em;
  color: #ffffff !important;
}

div.tabs li:hover {
  background: #333333;
}

div.tabs li.current:hover {
  background: #BB2222 !important;
}

div.tabs li.current {
  background: #BB2222 !important;
}

div.tabs li {
  border-right: 1px solid #333333;
}

div.tabs ul {
  display: inline;
  font-size: 100%;
  padding-top: 0em;
}

/* I want the menus to display directly below the Trac menu. */
#content {
  padding-top: 0px;
  margin-top: 0px;
}

div.tabs {
  margin-bottom: 0px;
  background-image: none;
}

div.nav {
  border-bottom: 1px solid #808080;
}

/*** Fix the weird size of the menus */
#mainnav {
  font-size: 100% !important;
}

div#main div.nav {
  min-height: 1em !important; /* We must have the right height for the menus. */
  border-bottom: 1px solid #333333; /* The plugin was giving a blue border. */
}
